/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.engine3d.entities.extended;

import com.jogamp.newt.event.InputEvent;
import com.jogamp.newt.event.MouseEvent;
import com.jogamp.opengl.GL;
import java.util.List;
import org.joml.Matrix4f;
import org.joml.Vector2i;
import org.joml.Vector4f;
import fr.amap.lidar.amapvox.engine3d.entities.base.Component;
import fr.amap.lidar.amapvox.engine3d.resources.postprocesses.base.FrameBuffer;
import fr.amap.lidar.amapvox.engine3d.resources.postprocesses.base.PostProcess;
import fr.amap.lidar.amapvox.engine3d.util.InputStatus;
import java.util.Collections;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.function.Consumer;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;

/**
 * Camera entity that can be placed in a scene to display other renderer
 * objects.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
@ThreadSafe
public final class Camera extends Component {

    /**
     * Unmodifiable global default FOV value (used by the voxel space component)
     */
    public static final float DEFAULT_FOV = 60.0f;
    
    /**
     * The field of view (in radians).
     */
    @GuardedBy("this")
    private float fov;

    /**
     * The near clip plane position in the z axis.
     */
    @GuardedBy("this")
    private float near;

    /**
     * The far clip plane position in the z axis.
     */
    @GuardedBy("this")
    private float far;

    /**
     * The screen aspect used to create th projection matrix.
     */
    @GuardedBy("this")
    private float aspect;

    /**
     * The screen proportion where this camera will render. X and Y = the top
     * left anchor position product of the current screen size. Z and X = the
     * width and height product of the current screen size.
     */
    @GuardedBy("this")
    private Vector4f proportions;

    /**
     * The screen dimensions (x = width, y = height).
     */
    @GuardedBy("this")
    private Vector2i dimensions;

    /**
     * The projection matrix.
     */
    @GuardedBy("this")
    private Matrix4f projection;

    /**
     * The framebuffer associated with this camera.
     */
    @GuardedBy("this")
    private FrameBuffer frameBuffer;

    /**
     * The post process list of this camera.
     */
    private final CopyOnWriteArrayList<PostProcess> postProcesses;
    
    /**
     * The rendered layouts.
     */
    private final CopyOnWriteArrayList<String> renderedLayouts;
    
    /**
     * If the zoom functionnality is enabled.
     */
    @GuardedBy("this")
    private boolean zoomEnabled;
    
    /**
     * The zooming callback function.
     */
    @GuardedBy("this")
    private Consumer<Float> onZoomCallback;

    /**
     * Create a camera with a FOV of 60, a near of 0.01 and a far of 1000.
     */
    public Camera() {
        this.fov = DEFAULT_FOV;
        this.near = 0.01f;
        this.far = 1000;
        this.aspect = 1f;
        this.proportions = new Vector4f(0, 0, 1, 1);
        this.dimensions = new Vector2i();
        this.projection = new Matrix4f().identity();
        this.postProcesses = new CopyOnWriteArrayList();
        this.frameBuffer = new FrameBuffer("CameraFrameBuffer", 500, 500);
        this.renderedLayouts = new CopyOnWriteArrayList();
    }

    /**
     * Add a new post process to this camera.
     * @param index
     * @param postProcess 
     */
    public final void addPostProcess(final int index, final PostProcess postProcess) {
        if (index < postProcesses.size()) {
            this.postProcesses.add(index, postProcess);
        } else {
            this.postProcesses.add(postProcess);
        }
    }
    
    /**
     * Add a rendered layout to this camera.
     * @param layout 
     */
    public final void addLayout(final String layout) {
        renderedLayouts.add(layout);
    }

    /**
     * Change the field of view.
     * @param fov New field of view (in radians).
     */
    public final synchronized void setFov(final float fov) {
        this.fov = fov;
    }

    /**
     * Set the zooming callback method.
     * @param onZoomCallback 
     */
    public final synchronized void setOnZoomCallback(final Consumer<Float> onZoomCallback) {
        this.onZoomCallback = onZoomCallback;
    }

    /**
     * Change the near clip plane position.
     * @param near
     */
    public final synchronized void setNear(final float near) {
        this.near = near;
    }

    /**
     * Change the far clip plane position.
     * @param far
     */
    public final synchronized void setFar(final float far) {
        this.far = far;
    }

    /**
     * Change he screen proportion where this camera will render. X and Y = the
     * top left anchor position product of the current screen size. Z and X =
     * the width and height product of the current screen size.
     * @param proportions
     */
    public final synchronized void setProportions(final Vector4f proportions) {
        this.proportions = proportions;
    }

    /**
     * Chnage if the zoom is enabled or not.
     * @param zoomEnabled 
     */
    public final synchronized void setZoomEnabled(final boolean zoomEnabled) {
        this.zoomEnabled = zoomEnabled;
    }

    /**
     * Get the projection matrix.
     * @return
     */
    public final synchronized Matrix4f getProjection() {
        return projection;
    }

    /**
     * Get the current rendered layouts.
     * @return 
     */
    public final List<String> getRenderedLayouts() {
        /* Return a read only copy of the layout list to make it thread safe */
        return Collections.unmodifiableList(renderedLayouts);
    }

    @Override
    public final void onInit(final GL gl) {
        /* Nothing needed to do */
    }

    @Override
    public void onInputting(final InputEvent event, final InputStatus status) {
        /* Perform zoom if that is enabled */
        if (status == InputStatus.MOUSE_WHEEL_MOVED && zoomEnabled) {
            final MouseEvent mouseEvent = (MouseEvent) event;
            fov -= mouseEvent.getRotation()[1];
            projection = new Matrix4f().setPerspective(
                    (float) Math.toRadians(fov), aspect, near, far);
            getParent().setModified(true);
            
            if (onZoomCallback != null) {
                onZoomCallback.accept(fov);
            }
        }
    }

    @Override
    public final void onUpdate(final GL gl, final float delta) {
        /* Nothing needed to do */
    }

    @Override
    public final void onDisplay(final GL gl) {
        /* Nothing needed to do */
    }
    
    public final synchronized void bind(final GL gl) {
        /* Set the drawing area to the current proportions of the screen */
        gl.glViewport(
                Math.round(proportions.x * (float) dimensions.x),
                Math.round(proportions.y * (float) dimensions.y),
                Math.round(proportions.z * (float) dimensions.x),
                Math.round(proportions.w * (float) dimensions.y)
        );
                
        /* Bind or unbind the framebuffer */
        if (frameBuffer != null && !postProcesses.isEmpty()) {
            /* Bind the current framebuffer */
            frameBuffer.bind(gl);
            
            /* Clear the color of the framebuffer by a transparent one */
            gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
            gl.glClear(GL.GL_COLOR_BUFFER_BIT);
        } else {
            /* Set the default framebuffer if there is no one post-process */
            gl.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0);
        }
        
        /* Clear the depth buffer */
        gl.glClear(GL.GL_DEPTH_BUFFER_BIT);
        
        /* Enable the depth test */
        gl.glEnable(GL.GL_DEPTH_TEST);
        
        /* Enable blending to mix multiple framebuffers (never disabled) */
        gl.glEnable(GL.GL_BLEND);
        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);  
    }

    public final synchronized void displayPostProcesses(final GL gl) {
        /* Draw each post process by sending the last texture */
        int inputTexture = frameBuffer.getColorTexture();

        for (int i = 0; i < postProcesses.size(); i++) {
            /* If the process is the last one, set the default framebuffer */
            if (i == postProcesses.size() - 1) {
                postProcesses.get(i).setFinalProcess(true);
            }

            /* Draw the post process by sending the last texture */
            postProcesses.get(i).setInputTexture(inputTexture);
            postProcesses.get(i).bind(gl);
            postProcesses.get(i).draw(gl);
            inputTexture = postProcesses.get(i).getOutputTexture();
        }
    }

    @Override
    public final void onRescale(final GL gl, final int width, final int height) {
        /* Get the screen aspect */
        aspect = (float) width / height;

        /* Update the projection matrix */
        projection = new Matrix4f().setPerspective(
                (float) Math.toRadians(fov), aspect, near, far);

        /* Update the screen dimensions */
        dimensions = new Vector2i(width, height);
        
        /* Resize the framebuffer */
        if (frameBuffer != null) {
            frameBuffer.unload(gl);
        }
        
        frameBuffer = new FrameBuffer("CameraFrameBuffer", width, height);

        /* Resize the post processes */
        for (PostProcess postProcess : postProcesses) {
            postProcess.setSize(gl, width, height);
        }
    }

    @Override
    public final void onDestroy(final GL gl) {
        /* Nothing needed to do */
    }

    @Override
    protected final synchronized void receiveMessage(
            final String name, final Object object) {
        /* Nothing needed to do */
    }

}
