/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.engine3d.entities.extended;

import org.joml.Matrix4f;
import org.joml.Vector3f;
import fr.amap.lidar.amapvox.engine3d.entities.base.Component;
import fr.amap.lidar.amapvox.engine3d.resources.materials.Material;
import fr.amap.lidar.amapvox.engine3d.resources.materials.StandardMaterial;
import fr.amap.lidar.amapvox.engine3d.resources.meshes.Mesh;
import fr.amap.lidar.amapvox.engine3d.resources.programs.Program;
import fr.amap.lidar.amapvox.engine3d.resources.textures.TextureCubemap;
import net.jcip.annotations.GuardedBy;

/**
 * A particular component to draw something in a scene.
 */
public abstract class Renderer extends Component {

    /**
     * The current used material.
     */
    @GuardedBy(value = "this")
    private Material material;

    /**
     * The mesh to draw.
     */
    @GuardedBy(value = "this")
    private Mesh mesh;

    /**
     * Create a renderer component.
     */
    public Renderer() {
        this.material = null;
        this.mesh = null;
    }

    /**
     * Change the current material.
     * @param material
     */
    public final synchronized void setMaterial(final Material material) {
        this.material = material;
    }

    /**
     * Change the current mesh.
     * @param mesh
     */
    public final synchronized void setMesh(final Mesh mesh) {
        this.mesh = mesh;
    }

    /**
     * Get the current material.
     * @return
     */
    public final synchronized Material getMaterial() {
        return material;
    }

    /**
     * Get the current mesh.
     * @return
     */
    public final synchronized Mesh getMesh() {
        return mesh;
    }

    /**
     * Generate a program of this renderer with its material and mesh.
     * @return
     */
    public final synchronized Program getProgram() {
        if (material == null || mesh == null) {
            throw (new NullPointerException("The current material or mesh is empty."));
        }

        final Program program = material.generateProgram();
        program.setMesh(mesh);
        return program;
    }

    @Override
    protected final synchronized void receiveMessage(final String name, final Object object) {
        if (null != name) {
            switch (name) {
                case "viewMatrix":
                    material.setView((Matrix4f) object);
                    break;
                case "projectionMatrix":
                    material.setProjection((Matrix4f) object);
                    break;
                case "viewPosition":
                    material.setViewPosition((Vector3f) object);
                    break;
                case "background":
                    if (material instanceof StandardMaterial) {
                        ((StandardMaterial) material).setCubemap((TextureCubemap) object);
                    }
                    break;
            }
        }
    }

}
