/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.engine3d.resources.meshes;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL4bc;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;
import org.apache.commons.lang3.RandomStringUtils;

/**
 * The mesh class is combined with a geometry to display it with rendering
 * options.
 */
@ThreadSafe
public final class Mesh {

    /**
     * The geometry object of this mesh.
     */
    @GuardedBy("this")
    private Geometry geometry;

    /**
     * Which shape to draw the primitive.
     */
    @GuardedBy("this")
    private DrawingType drawingType;

    /**
     * If the back faces are hiden or not (by default true).
     */
    @GuardedBy("this")
    private boolean faceCulling;

    /**
     * The depth function used with the depth buffer.
     */
    @GuardedBy("this")
    private DepthFunction depthFunction;

    /**
     * If the depth mask is used or not (by default true).
     */
    @GuardedBy("this")
    private boolean depthMask;

    /**
     * Random identifier of this mesh.
     */
    private final String hashtag;

    /**
     * Create a new mesh with en empty geometry, a drawing mode setted as
     * "triangles", one instance, face culling and depth mask enabled and depth
     * function setted as "less".
     */
    public Mesh() {
        this.geometry = null;
        this.drawingType = DrawingType.TRIANGLES;
        this.faceCulling = true;
        this.depthMask = true;
        this.depthFunction = DepthFunction.LESS;
        this.hashtag = RandomStringUtils.random(25);
    }

    /**
     * Chnage the current geometry.
     * @param geometry
     */
    public final synchronized void setGeometry(final Geometry geometry) {
        this.geometry = geometry;
    }

    /**
     * Chnage the current drawing type.
     * @param drawingType
     */
    public final synchronized void setDrawingType(final DrawingType drawingType) {
        this.drawingType = drawingType;
    }

    /**
     * If the back faces are hiden or not (by default true).
     * @param faceCulling
     */
    public final synchronized void setFaceCulling(final boolean faceCulling) {
        this.faceCulling = faceCulling;
    }

    /**
     * Set the depth function used with the depth buffer.
     * @param depthFunction
     */
    public final synchronized void setDepthFunction(final DepthFunction depthFunction) {
        this.depthFunction = depthFunction;
    }

    /**
     * If the depth mask is used or not (by default true)
     * @param depthMask
     */
    public final synchronized void setDepthMask(final boolean depthMask) {
        this.depthMask = depthMask;
    }

    /**
     * Get random identifier of this mesh.
     * @return
     */
    public final String getHashtag() {
        return hashtag;
    }

    /**
     * Generate and bind a vertex array with all data buffers.
     * @param gl
     * @return
     */
    public final int generateVerticesBuffer(final GL gl) {
        /* Create vertex buffer object */
        final IntBuffer vbo = Buffers.newDirectIntBuffer(1);

        /* Generate the vertex buffer objects */
        gl.glGenBuffers(vbo.limit(), vbo);

        /* Enable vertex position attribute */
        gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vbo.get(0));
        gl.getGL4().glEnableVertexAttribArray(0);
        gl.getGL4().glVertexAttribPointer(0, 3, GL.GL_FLOAT, false, 0, 0);

        /* Create vertex position buffers */
        gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vbo.get(0));
        final FloatBuffer vertexBuffer
                = Buffers.newDirectFloatBuffer(geometry.getVertices());
        gl.glBufferData(GL.GL_ARRAY_BUFFER,
                vertexBuffer.limit() * Float.BYTES,
                vertexBuffer, GL.GL_STATIC_DRAW);

        return vbo.get(0);
    }

    /**
     * Generate and bind a vertex array with all data buffers.
     * @param gl
     * @return
     */
    public final int generateNormalsBuffer(final GL gl) {
        /* Create vertex buffer object */
        final IntBuffer vbo = Buffers.newDirectIntBuffer(1);

        /* Generate the vertex buffer objects */
        gl.glGenBuffers(vbo.limit(), vbo);

        /* Create vertex normal buffers */
        gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vbo.get(0));
        final FloatBuffer normalBuffer
                = Buffers.newDirectFloatBuffer(geometry.getNormals());
        gl.glBufferData(GL.GL_ARRAY_BUFFER,
                normalBuffer.limit() * Float.BYTES,
                normalBuffer, GL.GL_STATIC_DRAW);

        /* Enable vertex normal attribute */
        gl.getGL4().glBindBuffer(GL.GL_ARRAY_BUFFER, vbo.get(0));
        gl.getGL4().glEnableVertexAttribArray(1);
        gl.getGL4().glVertexAttribPointer(1, 3, GL.GL_FLOAT, false, 0, 0);

        return vbo.get(0);
    }

    /**
     * Generate and bind a vertex array with all data buffers.
     * @param gl
     * @return
     */
    public final int generateTexcoordsBuffer(final GL gl) {
        /* Create vertex buffer object */
        final IntBuffer vbo = Buffers.newDirectIntBuffer(1);

        /* Generate the vertex buffer objects */
        gl.glGenBuffers(vbo.limit(), vbo);
        
        /* Create vertex texcoords buffers */
        gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vbo.get(0));
        final FloatBuffer texcoordBuffer
                = Buffers.newDirectFloatBuffer(geometry.getTexcoords());
        gl.glBufferData(GL.GL_ARRAY_BUFFER,
                texcoordBuffer.limit() * Float.BYTES,
                texcoordBuffer, GL.GL_STATIC_DRAW);

        /* Enable vertex texcoord attribute */
        gl.getGL4().glBindBuffer(GL.GL_ARRAY_BUFFER, vbo.get(0));
        gl.getGL4().glEnableVertexAttribArray(2);
        gl.getGL4().glVertexAttribPointer(2, 2, GL.GL_FLOAT, false, 0, 0);

        return vbo.get(0);
    }

    /**
     * Draw this mesh by follow these steps :
     * <ol>
     * <li>Bind the geometry.</li>
     * <li>Update the face culling.</li>
     * <li>Enable the depth test.</li>
     * <li>Update the face depth mask.</li>
     * <li>Draw each geometry instance with the current rendering method.</li>
     * <li>Unbind the geometry.</li>
     * </ol>
     * @param gl The current OpenGL interface
     * @param instances
     */
    public final synchronized void draw(final GL gl, final int instances) {
        if (geometry != null) {
            /* Hide back faces if face culling is enabled */
            if (faceCulling) {
                gl.glCullFace(GL.GL_BACK);
                gl.glEnable(GL.GL_CULL_FACE);
            } else {
                gl.glDisable(GL.GL_CULL_FACE);
            }

            /* Change the depth function */
            switch (depthFunction) {
                case LEQUAL:
                    gl.glDepthFunc(GL.GL_LEQUAL);
                    break;
                case LESS:
                    gl.glDepthFunc(GL.GL_LESS);
                    break;
            }

            /* Enabled blending */
            gl.glEnable(GL.GL_BLEND);

            /* Change depth mask */
            gl.glDepthMask(depthMask);

            /* Check if the drawing mode is not null */
            if (null != drawingType) {
                /* Get the vertex indices */
                final IntBuffer indexBuffer
                        = Buffers.newDirectIntBuffer(geometry.getIndices());

                /* Draw the available and selected drawing mode */
                switch (drawingType) {
                    case POINTS:
                        gl.getGL4().glDrawArraysInstanced(GL.GL_POINTS,
                                0, indexBuffer.limit(), instances);
                        break;
                    case LINES:
                        gl.getGL4().glDrawArraysInstanced(GL.GL_LINES,
                                0, indexBuffer.limit(), instances);
                        gl.getGL4().glDrawArraysInstanced(
                                GL4bc.GL_LINES_ADJACENCY, 0,
                                indexBuffer.limit(), instances);
                        break;
                    case STRIP_LINES:
                        gl.getGL4().glDrawArraysInstanced(
                                GL.GL_LINE_STRIP, 0, indexBuffer.limit(),
                                instances);
                        break;
                    case TRIANGLES:
                        gl.getGL4().glDrawArraysInstanced(
                                GL.GL_TRIANGLES, 0, indexBuffer.limit(),
                                instances);
                        break;
                }
            }
        }
    }

}
